/** @type {import('tailwindcss').Config} */
module.exports = {
  content: ["./docs/**/*.{html,js}"],
  theme: {
    extend: {
      colors: {
        'marvin-blue': 'rgba(24, 21, 68)',
      },
    },
  },
}